{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "f7af69c6",
   "metadata": {},
   "source": [
    "# Solution: the biggest earthquake in the UK this century\n",
    "\n",
    "## Import modules\n",
    "\n",
    "We first import the modules we will need to get and plot the data. We will use the `requests` module to query the USGS earthquake catalog, the `math` module to do some coordinate conversion and the `IPython` module to display a map image."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b2ca7665",
   "metadata": {
    "jupyter": {
     "outputs_hidden": false
    }
   },
   "outputs": [],
   "source": [
    "import math\n",
    "\n",
    "# sending requests to the web is not fully supported on jupyterlite yet, and the\n",
    "# cells below might error out on the browser (jupyterlite) version of this notebook\n",
    "import requests\n",
    "import IPython\n",
    "from IPython.display import Image"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "5af7fc3b",
   "metadata": {},
   "source": [
    "## Download the data\n",
    "\n",
    "We can reuse the code provided in the exercise notebook to query the USGS earthquake catalog API using the `requests` module."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "2fb11238",
   "metadata": {},
   "outputs": [],
   "source": [
    "earthquake_catalog_api_url = \"http://earthquake.usgs.gov/fdsnws/event/1/query\"\n",
    "query_parameters = {\n",
    "    \"format\": \"geojson\",\n",
    "    \"starttime\": \"2001-01-01\",\n",
    "    \"maxlatitude\": \"60.830\",\n",
    "    \"minlatitude\": \"49.877\",\n",
    "    \"maxlongitude\": \"1.767\",\n",
    "    \"minlongitude\": \"-8.182\",\n",
    "    \"minmagnitude\": \"1\",\n",
    "    \"orderby\": \"time-asc\"\n",
    "}\n",
    "quakes_response = requests.get(earthquake_catalog_api_url, params=query_parameters)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "0b12af96",
   "metadata": {},
   "source": [
    "We can check that the returned object is a `Response` as expected"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "df12b922",
   "metadata": {
    "jupyter": {
     "outputs_hidden": false
    }
   },
   "outputs": [],
   "source": [
    "type(quakes_response)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "78483adc",
   "metadata": {},
   "source": [
    "It is also useful to check whether the status of the returned response corresponds to there not being any client or server errors, with this being indicated by a status code of 200"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "301d0f95",
   "metadata": {
    "jupyter": {
     "outputs_hidden": false
    }
   },
   "outputs": [],
   "source": [
    "assert quakes_response.status_code == 200"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "18089dfc",
   "metadata": {},
   "source": [
    "## Parse the data as JSON"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "001f87a2",
   "metadata": {},
   "source": [
    "We saw in the exercise notebooks that the `Reponse` objects returned by `requests.get` have various attributes that allow accessing the response content in different formats including `Response.content` to get the raw `bytes` content and `Response.text` to get the response as a (Unicode) `str` object. We can print out all of the attributes of an object in Python using the inbuilt `dir` function; typically these will include attributes intended for internal use only which are conventionally indicated by prefixing with an underscore character `_`. We can display all the attributes without an initial underscore character as follows."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "8a9ff715",
   "metadata": {},
   "outputs": [],
   "source": [
    "[attribute for attribute in dir(quakes_response) if attribute[0] != '_']"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b70b30ed",
   "metadata": {},
   "source": [
    "As well as the `content`, `ok`, `status_code` and `text` attributes we already encountered, we can see there is also a `json` attribute, which seems like it could be  relevant to our task of decoding the response as JSON. We can find out more about this attribute by using a useful feature of Jupyter / IPython - by adding a question mark `?` to the end of a Python object the documentation string (`docstring`) for that object will be displayed."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "eb5c04d5",
   "metadata": {},
   "outputs": [],
   "source": [
    "quakes_response.json?"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "25bb29da",
   "metadata": {},
   "source": [
    "From this we can see that `quakes.response` is a method (function bound to an object) which returns a Python object corresponding to a JSON encoded response, which is exactly what we need. There are no required arguments so we can call the method by just adding a pair of empty parentheses."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d3fe9083",
   "metadata": {
    "jupyter": {
     "outputs_hidden": false
    }
   },
   "outputs": [],
   "source": [
    "quakes_json = quakes_response.json()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "bb65f060",
   "metadata": {},
   "source": [
    "If we had not been aware of the `json` method an alternative would be to use the `json` module directly as we saw previously. For example the following would give an equivalent result to the above.\n",
    "\n",
    "```Python\n",
    "import json\n",
    "quakes_json = json.loads(quakes_response.text)\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "50e7c774",
   "metadata": {},
   "source": [
    "## Investigate the data to discover how it is structured.\n",
    "\n",
    "Now that we have queried and decoded the data into a Python object, we need to do some exploration to find out how it is structure. In some cases there may be documentation we can use to help guide our exploration of data - for example [this page on the USGS earthquake catalog website](https://earthquake.usgs.gov/earthquakes/feed/v1.0/geojson.php) gives a summary of the GeoJSON format. However, in other cases we may not be so lucky and have to work with data with an undocumented format so it is also useful to consider how we might explore that data structure ourselves.  \n",
    "\n",
    "\n",
    "A potentially useful first step is to check what the type of the `quakes_json` object is."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "db9b83b3",
   "metadata": {
    "jupyter": {
     "outputs_hidden": false
    }
   },
   "outputs": [],
   "source": [
    "type(quakes_json)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "1b399e97",
   "metadata": {},
   "source": [
    "We see that `quakes_json` is a Python `dict` (dictionary) object. We might want to find out what keys are defined in this dictionary - we can do this by calling the `keys` method."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "dac578ed",
   "metadata": {
    "jupyter": {
     "outputs_hidden": false
    }
   },
   "outputs": [],
   "source": [
    "quakes_json.keys()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "bca63901",
   "metadata": {},
   "source": [
    "We see that the dictionary has three keys, all of which are strings. The `features` key in particular here looks potentially promising for our goal of finding the maximum magnitude event in the data (on the rationale that the magnitude is a *feature* of the event). We can check what type the value associated with the `features` key is."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3f8cfc3b",
   "metadata": {
    "jupyter": {
     "outputs_hidden": false
    }
   },
   "outputs": [],
   "source": [
    "type(quakes_json['features'])"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "20d84ce3",
   "metadata": {},
   "source": [
    "We find out that the `features` key contains a list. We can check the length of this list."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ff443f6f",
   "metadata": {
    "jupyter": {
     "outputs_hidden": false
    }
   },
   "outputs": [],
   "source": [
    "len(quakes_json['features'])"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "18fcdb4f",
   "metadata": {},
   "source": [
    "We could also use a set (which we encountered previously in the lesson on dictionaries) to find out what the set of types of all of the elements in the `quakes_json['features']` list is. Similarly to the list comprehensions we encountered in a previous lesson we can use a similar syntax here to succinctly construct the set we required."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "438ceb8a",
   "metadata": {},
   "outputs": [],
   "source": [
    "{type(feature) for feature in quakes_json['features']}"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "51c7d4fd",
   "metadata": {},
   "source": [
    "From this we see that all of the elements in the `quakes_json['features']` share the same Python `dict` type. We can use a similar approach to find out what keys all these dictionary objects have."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "154b37b9",
   "metadata": {},
   "outputs": [],
   "source": [
    "{tuple(feature.keys()) for feature in quakes_json['features']}"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "90e30a3d",
   "metadata": {},
   "source": [
    "This tells us that as well as all the elements being dictionaries, all of the dictionaries have the same keys. This suggests the list corresponds to a representation of a sequence of objects of the same type, with each dictionary containing the 'features' for a particular object, with in in this case the objects in questions being particular earthquake events.\n",
    "\n",
    "To check this idea, we can look at the value of a particular element in the `features` list - as we know all elements are dictionaries with the same keys, its reasonable to assume the first element `quakes_json['features'][0]` will be representative of the all the other elements in the list. We can start by summarising the keys and types of the values in this dictionary."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "70ecc3fc",
   "metadata": {},
   "outputs": [],
   "source": [
    "for key, value in quakes_json['features'][0].items():\n",
    "    print(key, type(value).__name__)       "
   ]
  },
  {
   "cell_type": "markdown",
   "id": "731c5488",
   "metadata": {},
   "source": [
    "We can also view the dictionary directly"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "8200d8dc",
   "metadata": {},
   "outputs": [],
   "source": [
    "quakes_json['features'][0]"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "d6239b3d",
   "metadata": {},
   "source": [
    "From this we can see the `properties` and `geometry` keys both themselves map to `dict` objects. Within these inner dictionaries are several keys which look relevant to our task of identifying the highest magnitude earthquake event and displaying its location on a map. Specifically the `mag` key in the `properties` dictionary seems likely to represent the magnitude of the event "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "4945c4e4",
   "metadata": {},
   "outputs": [],
   "source": [
    "quakes_json['features'][0]['properties']['mag']"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "06169bde",
   "metadata": {},
   "source": [
    "while the `coordinates` key in the `geometry` dictionary seems to represent the location of the event. "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "f13486e1",
   "metadata": {},
   "outputs": [],
   "source": [
    "quakes_json['features'][0]['geometry']['coordinates']"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "f3272dbf",
   "metadata": {},
   "source": [
    "If go to the URL listed as the value for the `url` key in the `properties` dictionary, "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "25cffdee",
   "metadata": {},
   "outputs": [],
   "source": [
    "quakes_json['features'][0]['properties']['url']"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "787fb52b",
   "metadata": {},
   "source": [
    "we can confirm that this is indeed a correct interpretation of the data as the listed magnitude corresponds to the value for the `mag` key while the longitude (East-West axis) and latitude (North-South axis) coordinates (in degrees) of the event location correspond to the first two elements respectively in the list associated with the `coordinates` key (with the third coordinate corresponding to the depth of the event)."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "5a0e2eb9",
   "metadata": {},
   "source": [
    "## Find the largest quake(s)\n",
    "\n",
    "Now that we have a handle on the structure of the data, we are ready to search through the data to identify the largest magnitude earthquake event(s). We are interested in finding the element (or elements) in a sequence which maximises some property - this operation is termed the [$\\arg\\max$ in mathematics and computer science](https://en.wikipedia.org/wiki/Arg_max). While there is built-in `max` function in Python there is no corresponding `argmax` function, though several external libraries including the NumPy library which we encounter in a subsequent lesson do include an `argmax` function.\n",
    "\n",
    "As a first approach, we will set the first eartquake to be the largest and as we iterate over the others we will replace the first one with whatever is larger."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "1c458ccd",
   "metadata": {},
   "outputs": [],
   "source": [
    "quakes = quakes_json['features']\n",
    "\n",
    "largest_so_far = quakes[0]\n",
    "for quake in quakes:\n",
    "    if quake['properties']['mag'] > largest_so_far['properties']['mag']:\n",
    "        largest_so_far = quake\n",
    "largest_so_far['properties']['mag']"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "73472ffd",
   "metadata": {},
   "source": [
    "Great, this gives us the largest earthquake... but, what if there were multiple earthquakes with the same magnitude? \n",
    "Can we know it from the above result?\n",
    "We will therefore loop over all of the event details in the `features` list and construct a list of the event or events for which the magnitude is currently the largest, creating a new list if the magnitude of the current event is larger than the previous largest or adding the event to the previous list if it has an equal magnitude. After iterating through all the events this list should contain the details of the event(s) with the largest magnitude. An example implementation of this approach is as follows."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "f88b9783",
   "metadata": {
    "jupyter": {
     "outputs_hidden": false
    }
   },
   "outputs": [],
   "source": [
    "largest_magnitude_events = [quakes[0]]\n",
    "\n",
    "for quake in quakes:\n",
    "    if quake['properties']['mag'] > largest_magnitude_events[0]['properties']['mag']:\n",
    "        largest_magnitude_events = [quake]\n",
    "    elif quake['properties']['mag'] == largest_magnitude_events[0]['properties']['mag']:\n",
    "        largest_magnitude_events += [quake]"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "fd5331fe",
   "metadata": {},
   "source": [
    "We can now check if there was a single event with the maximum magnitude or multiple"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "82150ad8",
   "metadata": {},
   "outputs": [],
   "source": [
    "len(largest_magnitude_events)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "cda38645",
   "metadata": {},
   "source": [
    "It turns out there are two events with the same maximal magnitude. As a sanity check we can print the magnitude of both events to check that they match"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e7a16a97",
   "metadata": {
    "jupyter": {
     "outputs_hidden": false
    }
   },
   "outputs": [],
   "source": [
    "print([quake['properties']['mag'] for quake in largest_magnitude_events])"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "1e8d957a",
   "metadata": {},
   "source": [
    "## Get a map at the point of the quake"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "d73fbf87",
   "metadata": {},
   "source": [
    "We saw something similar in the [Greengraph example](../ch01python/010exemplar.html#More-complex-functions) [(notebook version)](../ch01python/010exemplar.ipynb#More-complex-functions) of the previous chapter."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "766ce50b",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "def deg2num(lat_deg, lon_deg, zoom):\n",
    "    \"\"\"Convert latitude and longitude to XY tiles coordinates.\"\"\"\n",
    "\n",
    "    lat_rad = math.radians(lat_deg)\n",
    "    n = 2.0 ** zoom\n",
    "    x_tiles_coord = int((lon_deg + 180.0) / 360.0 * n)\n",
    "    y_tiles_coord = int((1.0 - math.asinh(math.tan(lat_rad)) / math.pi) / 2.0 * n)\n",
    "\n",
    "    return (x_tiles_coord, y_tiles_coord)\n",
    "\n",
    "def request_map_at(latitude, longitude, zoom=10, satellite=True):\n",
    "    \"\"\"Retrieve a map from Google at a given location.\"\"\"\n",
    "\n",
    "    base_url = \"https://mt0.google.com/vt?\"\n",
    "    x_coord, y_coord = deg2num(latitude, longitude, zoom)\n",
    "\n",
    "    params = dict(\n",
    "        x=x_coord,\n",
    "        y=y_coord,\n",
    "        z=zoom,\n",
    "    )\n",
    "    if satellite:\n",
    "        params['lyrs'] = 's'\n",
    "    \n",
    "    return requests.get(base_url, params=params)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7b123fdc",
   "metadata": {},
   "source": [
    "As a test we can check the map displayed for the coordinates of the [Prime meridian at the Royal Observatory Greenwich](https://geohack.toolforge.org/geohack.php?pagename=Prime_meridian_(Greenwich)&params=51_28_40.1_N_0_0_5.3_W_type:landmark_globe:earth_region:GB_scale:1000)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "bb1b386e",
   "metadata": {},
   "outputs": [],
   "source": [
    "map_response = request_map_at(51.477806, -0.001472, zoom=14)\n",
    "Image(map_response.content)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "395f7f91",
   "metadata": {},
   "source": [
    "## Display the maps\n",
    "\n",
    "We can now finally show the maps for the locations of the maximum magnitude earthquake events. As additional check we also print the description under the `title` key in the `properties` dictionary for the event to check it tallies with the location shown in the displayed map."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "5177b0a3",
   "metadata": {},
   "outputs": [],
   "source": [
    "for quake in largest_magnitude_events:\n",
    "    longitude = quake['geometry']['coordinates'][0]\n",
    "    latitude = quake['geometry']['coordinates'][1]\n",
    "    print(quake['properties']['title'])\n",
    "    display(Image(request_map_at(latitude, longitude, 12).content))"
   ]
  }
 ],
 "metadata": {
  "jekyll": {
   "display_name": "Quakes Solution"
  },
  "jupytext": {
   "main_language": "python",
   "notebook_metadata_filter": "-kernelspec,jupytext,jekyll"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}